@@ -1,4 +1,4 @@
-VERSION = "0.91"
+VERSION = "0.91.1"
 
 
 
@@ -90,6 +90,7 @@ def main():
                     ("Quit", "quit"),
                     ],
                 focus_buttons=True,
+                bind_enter_esc=False,
                 title="Finished")
 
             if result == "quit":
