@@ -1,6 +1,7 @@
 #! /usr/bin/env python
 # -*- coding: utf-8 -*-
 
+from __future__ import division
 import urwid
 import bdb
 from code import InteractiveConsole
@@ -38,6 +39,7 @@ Keys:
     ctrl-u/d - page up/down
     h/l - scroll left/right
     g/G - start/end
+    L - go to line
     / - search
     ,/. - search next/previous
 
@@ -48,6 +50,22 @@ Keys:
 
     f1/?/H - show this help screen
 
+Keys in variables list:
+
+    \ - expand/collapse
+    t/r/s - show type/repr/str for this variable
+    enter - edit options
+    h - toggle highlighting
+    w - toggle watching
+
+Keys in stack list:
+
+    enter - jump to frame
+
+Keys in breakpoints view:
+
+    enter - edit breakpoint
+
 License:
 --------
 
@@ -121,7 +139,7 @@ class Debugger(bdb.Bdb):
         stack, index = self.get_stack(frame, tb)
 
         for i, (s_frame, lineno) in enumerate(stack):
-            if s_frame is self.bottom_frame:
+            if s_frame is self.bottom_frame and index >= i:
                 stack = stack[i:]
                 index -= i
 
@@ -158,7 +176,8 @@ class Debugger(bdb.Bdb):
         self.ui.call_with_ui(self.ui.interaction, exc_tuple)
 
     def get_call_path(self):
-        return "/".join(str(id(frame.f_code)) for frame, lineno in self.stack)
+        return "/".join(str(id(frame.f_code)) 
+                for frame, lineno in self.stack[:self.curindex+1])
 
     def user_call(self, frame, argument_list):
         """This method is called when there is the remote possibility
@@ -217,6 +236,40 @@ class Debugger(bdb.Bdb):
 
 
 # UI stuff --------------------------------------------------------------------
+def make_canvas(txt, attr, maxcol, fill_attr=None):
+    processed_txt = []
+    processed_attr = []
+    processed_cs = []
+
+    for line, line_attr in zip(txt, attr):
+        # filter out zero-length attrs
+        line_attr = [(aname, l) for aname, l in line_attr if l > 0]
+
+        diff = maxcol - len(line)
+        if diff > 0:
+            line += " "*diff
+            line_attr.append((fill_attr, diff))
+        else:
+            from urwid.util import rle_subseg
+            line = line[:maxcol]
+            line_attr = rle_subseg(line_attr, 0, maxcol)
+        
+        from urwid.util import apply_target_encoding
+        line, line_cs = apply_target_encoding(line)
+
+        processed_txt.append(line)
+        processed_attr.append(line_attr)
+        processed_cs.append(line_cs)
+
+    return urwid.TextCanvas(
+            processed_txt, 
+            processed_attr, 
+            processed_cs, 
+            maxcol=maxcol)
+
+
+
+
 class MyConsole(InteractiveConsole):
     def __init__(self, locals):
         InteractiveConsole.__init__(self, locals)
@@ -291,51 +344,52 @@ class Variable(urwid.FlowWidget):
     def selectable(self):
         return True
 
-    def rows(self, (maxcol,), focus=False):
-        return 1
+    SIZE_LIMIT = 20
 
-    def render(self, (maxcol,), focus=False):
+    def rows(self, size, focus=False):
+        if (self.prefix is not None 
+                and self.var_label is not None 
+                and len(self.prefix) + len(self.var_label) > self.SIZE_LIMIT):
+            return 2
+        else:
+            return 1
+
+    def render(self, size, focus=False):
+        maxcol = size[0]
         if focus:
             apfx = "focused "+self.attr_prefix+" "
         else:
             apfx = self.attr_prefix+" "
 
-
         if self.value_str is not None:
             if self.var_label is not None:
-                text = self.prefix + self.var_label +": " + self.value_str
+                if len(self.prefix) + len(self.var_label) > self.SIZE_LIMIT:
+                    # label too long? generate separate value line
+                    text = [self.prefix + self.var_label,
+                            self.prefix+"  " + self.value_str]
 
-                attr = [
-                        (apfx+"label", len(self.prefix)+len(self.var_label)+2),
-                        (apfx+"value", len(self.value_str)),
-                        ]
+                    attr = [[(apfx+"label", len(self.prefix)+len(self.var_label))],
+                            [(apfx+"value", len(self.prefix)+2+len(self.value_str))]]
+                else:
+                    text = [self.prefix + self.var_label +": " + self.value_str]
+
+                    attr = [[
+                            (apfx+"label", len(self.prefix)+len(self.var_label)+2),
+                            (apfx+"value", len(self.value_str)),
+                            ]]
             else:
-                text = self.prefix + self.value_str
+                text = [self.prefix + self.value_str]
 
-                attr = [
+                attr = [[
                         (apfx+"label", len(self.prefix)),
                         (apfx+"value", len(self.value_str)),
-                        ]
+                        ]]
         else:
-            text = self.prefix + self.var_label
+            text = [self.prefix + self.var_label]
 
-            attr = [ (apfx+"label", len(self.prefix) + len(self.var_label)), ]
+            attr = [[ (apfx+"label", len(self.prefix) + len(self.var_label)), ]]
 
-        attr = [(aname, l) for aname, l in attr if l > 0]
-
-        diff = maxcol - len(text)
-        if diff > 0:
-            text += " "*diff
-            attr.append((apfx+"value", diff))
-        else:
-            from urwid.util import rle_subseg
-            text = text[:maxcol]
-            attr = rle_subseg(attr, 0, maxcol)
-        
-        from urwid.util import apply_target_encoding
-        txt, cs = apply_target_encoding(text)
-
-        return urwid.TextCanvas([txt], [attr], [cs], maxcol=maxcol) 
+        return make_canvas(text, attr, maxcol, apfx+"value")
 
     def keypress(self, size, key):
         return key
@@ -380,19 +434,7 @@ class StackFrame(urwid.FlowWidget):
         text += loc
         attr.append((apfx+"frame location", len(loc)))
 
-        diff = maxcol - len(text)
-        if diff > 0:
-            text += " "*diff
-            attr.append((apfx+"frame location", diff))
-        else:
-            from urwid.util import rle_subseg
-            text = text[:maxcol]
-            attr = rle_subseg(attr, 0, maxcol)
-        
-        from urwid.util import apply_target_encoding
-        txt, cs = apply_target_encoding(text)
-
-        return urwid.TextCanvas([txt], [attr], [cs], maxcol=maxcol) 
+        return make_canvas([text], [attr], maxcol, apfx+"frame location")
 
     def keypress(self, size, key):
         return key
@@ -567,6 +609,16 @@ class SearchBox(urwid.Edit):
 
 
 
+class InspectInfo(object):
+    def __init__(self):
+        self.show_detail = False
+        self.display_type = "type"
+        self.highlighted = False
+        self.watched = False
+
+
+
+
 class DebuggerUI(object):
     def __init__(self, dbg):
         self.debugger = dbg
@@ -574,7 +626,7 @@ class DebuggerUI(object):
 
         self.search_box = None
 
-        self.vars_with_detail = {}
+        self.inspect_info = {}
 
         self.source = urwid.SimpleListWalker([])
         self.source_list = urwid.ListBox(self.source)
@@ -613,8 +665,8 @@ class DebuggerUI(object):
 
         self.columns = urwid.Columns(
                     [
-                        self.lhs_col, 
-                        ("fixed", 40, self.rhs_col), 
+                        ("weight", 1, self.lhs_col), 
+                        ("weight", 0.5, self.rhs_col), 
                         ],
                     dividechars=1)
 
@@ -625,17 +677,84 @@ class DebuggerUI(object):
             header))
 
         # variable listeners --------------------------------------------------
-        def toggle_variable_detail(w, size, key):
-            cpath = self.debugger.get_call_path()
-            id_path_to_detail = self.vars_with_detail.setdefault(cpath, {})
-
+        def change_var_state(w, size, key):
             var, pos = self.var_list._w.get_focus()
-            id_path_to_detail[var.id_path] = not id_path_to_detail.get(
-                    var.id_path, False)
+
+            cpath = self.debugger.get_call_path()
+            id_path_to_iinfo = self.inspect_info.setdefault(cpath, {})
+            iinfo = id_path_to_iinfo.setdefault(var.id_path, InspectInfo())
+
+            if key == "\\": iinfo.show_detail = not iinfo.show_detail
+            elif key == "t": iinfo.display_type = "type"
+            elif key == "r": iinfo.display_type = "repr"
+            elif key == "s": iinfo.display_type = "str"
+            elif key == "h": iinfo.highlighted = not iinfo.highlighted
+            elif key == "w": iinfo.watched = not iinfo.watched
 
             self.set_locals(self.debugger.curframe.f_locals)
 
-        self.var_list.listen("enter", toggle_variable_detail)
+        def edit_variable_detail(w, size, key):
+            var, pos = self.var_list._w.get_focus()
+
+            cpath = self.debugger.get_call_path()
+            id_path_to_iinfo = self.inspect_info.setdefault(cpath, {})
+            iinfo = id_path_to_iinfo.setdefault(var.id_path, InspectInfo())
+
+            def make_lv(label, value):
+                return urwid.AttrWrap(urwid.Text([
+                    ("label", label), str(value)]),
+                    "fixed value", "fixed value")
+
+            rb_grp = []
+            rb_show_type = urwid.RadioButton(rb_grp, "Show Type", 
+                    iinfo.display_type == "type")
+            rb_show_repr = urwid.RadioButton(rb_grp, "Show repr()",
+                    iinfo.display_type == "repr")
+            rb_show_str = urwid.RadioButton(rb_grp, "Show str()",
+                    iinfo.display_type == "str")
+
+            expanded_checkbox = urwid.CheckBox("Expanded", iinfo.show_detail)
+            highlighted_checkbox = urwid.CheckBox("Highlighted", iinfo.highlighted)
+            watched_checkbox = urwid.CheckBox("Watched", iinfo.highlighted)
+
+            def make_lv(label, value):
+                return urwid.AttrWrap(urwid.Text([
+                    ("label", label), str(value)]),
+                    "fixed value", "fixed value")
+
+            lb = urwid.ListBox([
+                #make_lv("Call Path:       ", cpath),
+                make_lv("Identifier Path: ", var.id_path),
+                urwid.Text(""),
+                ]+rb_grp+[
+                urwid.Text(""),
+                expanded_checkbox,
+                highlighted_checkbox,
+                watched_checkbox,
+                ])
+
+            if self.dialog(lb, [
+                ("OK", True),
+                ("Cancel", False),
+                ], title="Variable Inspection Options"):
+                
+                iinfo.show_detail = expanded_checkbox.get_state()
+                iinfo.highlighted = highlighted_checkbox.get_state()
+                iinfo.watched = watched_checkbox.get_state()
+
+                if rb_show_type.get_state(): iinfo.display_type = "type"
+                elif rb_show_repr.get_state(): iinfo.display_type = "repr"
+                elif rb_show_str.get_state(): iinfo.display_type = "str"
+
+            self.set_locals(self.debugger.curframe.f_locals)
+
+        self.var_list.listen("\\", change_var_state)
+        self.var_list.listen("t", change_var_state)
+        self.var_list.listen("r", change_var_state)
+        self.var_list.listen("s", change_var_state)
+        self.var_list.listen("h", change_var_state)
+        self.var_list.listen("w", change_var_state)
+        self.var_list.listen("enter", edit_variable_detail)
 
         # stack listeners -----------------------------------------------------
         def examine_frame(w, size, key):
@@ -662,15 +781,20 @@ class DebuggerUI(object):
             enabled_checkbox = urwid.CheckBox(
                     "Enabled", bp.enabled)
             cond_edit = urwid.Edit([
-                ("label", "Condition: ")
+                ("label", "Condition:               ")
                 ], cond)
+            ign_count_edit = urwid.IntEdit([
+                ("label", "Ignore the next N times: ")
+                ], bp.ignore)
 
             lb = urwid.ListBox([
                 make_lv("File: ", bp.file),
                 make_lv("Line: ", bp.line),
                 make_lv("Hits: ", bp.hits),
+                urwid.Text(""),
                 enabled_checkbox,
-                urwid.AttrWrap(cond_edit, "value", "value")
+                urwid.AttrWrap(cond_edit, "value", "value"),
+                urwid.AttrWrap(ign_count_edit, "value", "value"),
                 ])
 
             result = self.dialog(lb, [
@@ -681,6 +805,7 @@ class DebuggerUI(object):
 
             if result == True:
                 bp.enabled = enabled_checkbox.get_state()
+                bp.ignore = int(ign_count_edit.value())
                 cond = cond_edit.get_edit_text()
                 if cond:
                     bp.cond = cond
@@ -792,17 +917,19 @@ class DebuggerUI(object):
                 self.rhs_col.set_focus(self.rhs_col.widget_list[subself.idx])
 
         def grow_sidebar(w, size, key):
-            _, width = self.columns.column_types[1]
-            width += 5
-            if width <= 75:
-                self.columns.column_types[1] = "fixed", width
+            _, weight = self.columns.column_types[1]
+
+            if weight < 5:
+                weight *= 1.25
+                self.columns.column_types[1] = "weight", weight
                 self.columns._invalidate()
 
         def shrink_sidebar(w, size, key):
-            _, width = self.columns.column_types[1]
-            width -= 5
-            if width >= 10:
-                self.columns.column_types[1] = "fixed", width
+            _, weight = self.columns.column_types[1]
+
+            if weight > 1/5:
+                weight /= 1.25
+                self.columns.column_types[1] = "weight", weight
                 self.columns._invalidate()
 
         def move_home(w, size, key):
@@ -811,6 +938,22 @@ class DebuggerUI(object):
         def move_end(w, size, key):
             self.source.set_focus(len(self.source))
 
+        def go_to_line(w, size, key):
+            _, line = self.source.get_focus()
+
+            lineno_edit = urwid.IntEdit([
+                ("label", "Line number: ")
+                ], line+1)
+
+            if self.dialog(
+                    urwid.ListBox([ urwid.AttrWrap(lineno_edit, "value") ]),
+                    [
+                        ("OK", True),
+                        ("Cancel", False),
+                        ], title="Go to Line Number"):
+                lineno = min(max(0, int(lineno_edit.value())-1), len(self.source)-1)
+                self.source.set_focus(lineno)
+
         def move_down(w, size, key):
             w.keypress(size, "down")
 
@@ -992,6 +1135,7 @@ class DebuggerUI(object):
         self.top.listen("end", move_end)
         self.top.listen("g", move_home)
         self.top.listen("G", move_end)
+        self.top.listen("L", go_to_line)
 
         self.top.listen("b", toggle_breakpoint)
         self.top.listen("m", pick_module)
@@ -1002,10 +1146,7 @@ class DebuggerUI(object):
         self.top.listen("?", help)
 
         # setup ---------------------------------------------------------------
-        try:
-            import urwid.curses_display as display
-        except ImportError:
-            import urwid.raw_display as display
+        import urwid.raw_display as display
 
         self.screen = display.Screen()
         self.setup_palette(self.screen)
@@ -1091,12 +1232,23 @@ class DebuggerUI(object):
             ("current breakpoint focused source", "white", "dark red"),
 
             ("variables", "black", "dark cyan"),
+            ("variable separator", "light gray", "dark cyan"),
 
             ("var label", "dark blue", "dark cyan"),
             ("var value", "black", "dark cyan"),
             ("focused var label", "dark blue", "dark green"),
             ("focused var value", "black", "dark green"),
 
+            ("highlighted var label", "white", "dark cyan"),
+            ("highlighted var value", "black", "dark cyan"),
+            ("focused highlighted var label", "white", "dark green"),
+            ("focused highlighted var value", "black", "dark green"),
+
+            ("return label", "white", "dark blue"),
+            ("return value", "black", "dark cyan"),
+            ("focused return label", "light gray", "dark blue"),
+            ("focused return value", "black", "dark green"),
+
             ("return label", "white", "dark blue"),
             ("return value", "black", "dark cyan"),
             ("focused return label", "light gray", "dark blue"),
@@ -1372,12 +1524,13 @@ class DebuggerUI(object):
 
     def set_locals(self, locals):
         vars = locals.keys()
-        vars.sort()
+        vars.sort(key=lambda n: n.lower())
 
+        watch_list = []
         loc_list = []
 
         cpath = self.debugger.get_call_path()
-        id_path_to_detail = self.vars_with_detail.get(cpath, {})
+        id_path_to_iinfo = self.inspect_info.get(cpath, {})
 
         try:
             import numpy
@@ -1385,27 +1538,47 @@ class DebuggerUI(object):
         except ImportError:
             HAVE_NUMPY = 0
 
-        def add_var(prefix, label, value, id_path=None, attr_prefix=None):
+        def add_var(prefix, var_label, value_str, id_path=None, attr_prefix=None):
+            iinfo = id_path_to_iinfo.get(id_path, InspectInfo())
+            if iinfo.highlighted:
+                attr_prefix = "highlighted var"
+
+            if iinfo.watched:
+                watch_list.append(Variable(prefix, var_label, value_str, id_path, attr_prefix))
+
+            loc_list.append(Variable(prefix, var_label, value_str, id_path, attr_prefix))
+
+        def display_var(prefix, label, value, id_path=None, attr_prefix=None):
             if id_path is None:
                 id_path = label
 
+            iinfo = id_path_to_iinfo.get(id_path, InspectInfo())
+
             if isinstance(value, (int, float, long, complex)):
-                loc_list.append(Variable(prefix, label, repr(value), id_path, attr_prefix))
+                add_var(prefix, label, repr(value), id_path, attr_prefix)
             elif isinstance(value, (str, unicode)):
-                loc_list.append(Variable(prefix, label, repr(value)[:200], id_path, attr_prefix))
+                add_var(prefix, label, repr(value)[:200], id_path, attr_prefix)
             elif isinstance(value, type):
-                loc_list.append(Variable(prefix, label, "type "+value.__name__, id_path, attr_prefix))
+                add_var(prefix, label, "type "+value.__name__, id_path, attr_prefix)
             else:
                 if isinstance(value, numpy.ndarray):
-                    loc_list.append(Variable(prefix, label, 
+                    add_var(prefix, label, 
                         "ndarray %s %s" % (value.dtype, value.shape), 
-                        id_path, attr_prefix))
+                        id_path, attr_prefix)
                 else:
-                    loc_list.append(Variable(prefix, label, type(value).__name__, id_path, attr_prefix))
+                    if iinfo.display_type == "type":
+                        displayed_value = type(value).__name__
+                    elif iinfo.display_type == "repr":
+                        displayed_value = repr(value)
+                    elif iinfo.display_type == "str":
+                        displayed_value = str(value)
+                    else:
+                        displayed_value = "ERROR: Invalid display_type"
 
-                do_detail = id_path_to_detail.get(id_path, False)
+                    add_var(prefix, label, 
+                        displayed_value, id_path, attr_prefix)
 
-                if not do_detail:
+                if not iinfo.show_detail:
                     return
 
                 # set ---------------------------------------------------------
@@ -1413,25 +1586,19 @@ class DebuggerUI(object):
                     for i, entry in enumerate(value):
                         if i % 10 == 0 and i: 
                             cont_id_path = "%s.cont-%d" % (id_path, i)
-                            if not id_path_to_detail.get(cont_id_path, False):
-                                loc_list.append(Variable(
-                                    prefix+"  ", "...", None, cont_id_path))
+                            if not id_path_to_iinfo.get(
+                                    cont_id_path, InspectInfo()).show_detail:
+                                add_var(prefix+"  ", "...", None, cont_id_path)
                                 break
 
-                        add_var(prefix+"  ", None, entry,
+                        display_var(prefix+"  ", None, entry,
                             "%s[%d]" % (id_path, i))
                     if not value:
-                        loc_list.append(Variable(prefix+"  ", "<empty>", None))
+                        add_var(prefix+"  ", "<empty>", None)
                     return
 
-                # dict --------------------------------------------------------
+                # containers --------------------------------------------------
                 key_it = None
-                try:
-                    key_it = value.iterkeys()
-                except:
-                    pass
-
-                # list, tuple -------------------------------------------------
                 try:
                     l = len(value)
                 except:
@@ -1439,42 +1606,68 @@ class DebuggerUI(object):
                 else:
                     key_it = xrange(l)
 
+                try:
+                    key_it = value.iterkeys()
+                except:
+                    pass
+
                 if key_it is not None:
                     cnt = 0
                     for key in key_it:
                         if cnt % 10 == 0 and cnt: 
                             cont_id_path = "%s.cont-%d" % (id_path, cnt)
-                            if not id_path_to_detail.get(cont_id_path, False):
-                                loc_list.append(Variable(
-                                    prefix+"  ", "...", None, cont_id_path))
+                            if not id_path_to_iinfo.get(
+                                    cont_id_path, InspectInfo()).show_detail:
+                                add_var(
+                                    prefix+"  ", "...", None, cont_id_path)
                                 break
 
-                        add_var(prefix+"  ", repr(key), value[key],
+                        display_var(prefix+"  ", repr(key), value[key],
                             "%s[%r]" % (id_path, key))
                         cnt += 1
                     if not cnt:
-                        loc_list.append(Variable(prefix+"  ", "<empty>", None))
+                        add_var(prefix+"  ", "<empty>", None)
                     return
 
                 # class types -------------------------------------------------
+                try:
+                    key_it = value.__slots__
+                except:
+                    pass
+                else:
+                    for key in key_it:
+                        if key[0] == "_":
+                            continue
+
+                        if hasattr(value, key):
+                            display_var(prefix+"  ", 
+                                    ".%s" % key, getattr(value, key), 
+                                    "%s.%s" % (id_path, key))
+
                 try:
                     key_it = value.__dict__.iterkeys()
                 except:
                     pass
                 else:
-                    for count, key in enumerate(key_it):
-                        add_var(prefix+"  ", 
-                                ".%s" % key, getattr(value, key), 
-                                "%s.%s" % (id_path, key))
-                    return
-
+                    for key in key_it:
+                        if key[0] != "_":
+                            display_var(prefix+"  ", 
+                                    ".%s" % key, getattr(value, key), 
+                                    "%s.%s" % (id_path, key))
 
         if "__return__" in vars:
-            add_var("", "Return", locals["__return__"], attr_prefix="return")
+            display_var("", "Return", locals["__return__"], attr_prefix="return")
 
         for var in vars:
             if not var[0] in "_.":
-                add_var("", var, locals[var])
+                display_var("", var, locals[var])
+
+        if watch_list:
+            loc_list = (watch_list 
+                    + [urwid.AttrWrap(
+                        urwid.Text("---", align="center"), 
+                        "variable separator")]
+                    + loc_list)
 
         self.locals[:] = loc_list
 
